from typing import Optional

import regex as re

from .common import BlockMatch

PROGRAM_INFO_START_RE = re.compile(
    r"""
(^[\* ]{14,}
\ PROGRAM\ (?P<key>
        (?:STARTED\ (?:AT|ON|BY))|
        PROCESS\ ID|
        )
    \s+
    (?P<value>.+)
    \n
    ){4}

[\* ]{14,}
\ PROGRAM\ (?P<inkey>(?:STARTED\ IN))
    \s+
    (?P<invalue>.+)
    \n
(\ \s{42}(?P<invalue>.+)\n)*
""",
    re.VERSION1 | re.VERBOSE | re.MULTILINE,
)


PROGRAM_INFO_STOP_RE = re.compile(
    r"""
(^[\* ]{14,}
\ PROGRAM\ (?P<key>
        ENDED\ AT|
        (?:RAN\ (?:ON|BY))|
        PROCESS\ ID|
        )
    \s+
    (?P<value>.+)
    \n
    ){4}

[\* ]{14,}
\ PROGRAM\ (?P<inkey>(?:STOPPED\ IN))
    \s+
    (?P<invalue>.+)
    \n
(\ \s{42}(?P<invalue>.+)\n)*
""",
    re.VERSION1 | re.VERBOSE | re.MULTILINE,
)


def match_program_info(content: str) -> Optional[BlockMatch]:
    spans = []
    match = PROGRAM_INFO_START_RE.search(content)

    if not match:
        return None

    keys = match.captures("key")
    values = match.captures("value")

    result = dict(zip([k.lower() for k in keys], values))
    result[match["inkey"].lower()] = "".join(match.captures("invalue"))
    spans += match.spans(0)

    match = PROGRAM_INFO_STOP_RE.search(content)

    if match:
        keys = match.captures("key")
        values = match.captures("value")

        result.update(zip([k.lower() for k in keys], values))
        result[match["inkey"].lower()] = "".join(match.captures("invalue"))
        spans += match.spans(0)

    return BlockMatch({"program info": result}, spans)



import contextlib
import mmap
import re
from collections.abc import Iterator

__all__ = ["parse", "parse_iter", "BlockIterator"]


@contextlib.contextmanager
def as_byteorstringlike(fh_or_content):
    """
    Yields a tuple (content, content_type),
    where content_type is True if content is a unicode string,
    and False if it is a byte-like (and needs to be decoded).
    """

    if isinstance(fh_or_content, str):
        yield fh_or_content, True
    elif isinstance(fh_or_content, bytes):
        yield fh_or_content, False
    else:
        # if the handle is a file handle, use mmap to return a byte-like object
        mmapped = mmap.mmap(fh_or_content.fileno(), 0, access=mmap.ACCESS_READ)

        try:
            yield mmapped, False
        finally:
            mmapped.close()


# MULTILINE and VERBOSE regex to match coordinate lines in a frame:
POS_MATCH_REGEX = r"""
^                                                               # Linestart
[ \t]*                                                          # Optional white space
(?P<sym>[A-Za-z]+[A-Za-z0-9]*)\s+                               # get the symbol
(?P<x> [\-\+]? (\d*\.\d+|\d+\.?\d*) ([Ee][\+\-]?\d+)? ) [ \t]+  # Get x
(?P<y> [\-\+]? (\d*\.\d+|\d+\.?\d*) ([Ee][\+\-]?\d+)? ) [ \t]+  # Get y
(?P<z> [\-\+]? (\d*\.\d+|\d+\.?\d*) ([Ee][\+\-]?\d+)? )         # Get z
"""

# MULTILINE and VERBOSE regex to match frames:
FRAME_MATCH_REGEX = r"""
                                        # First line contains an integer
                                        # and only an integer: the number of atoms
^[ \t]* (?P<natoms> [0-9]+) [ \t]*[\n]  # End first line
(?P<comment>.*) [\n]                    # The second line is a comment
(?P<positions>                          # This is the block of positions
    (
        (
            \s*                         # White space in front of the element spec is ok
            (
                [A-Za-z]+[A-Za-z0-9]*   # Element spec
                (
                   \s+                  # White space in front of the number
                   [\-\+]?              # Plus or minus in front of the number (optional)
                   (\d*                 # optional decimal in the beginning .0001 is ok, for example
                    \.                  # There has to be a dot followed by
                    \d+)                # at least one decimal
                    |                   # OR
                   (\d+                 # at least one decimal, followed by
                    \.?                 # an optional dot
                    \d*)                # followed by optional decimals
                   ([Ee][\+\-]?\d+)?    # optional exponents E+03, e-05
                ){3}                    # I expect three float values
                |
                \#                      # If a line is commented out, that is also ok
            )
            .*                          # I do not care what is after the comment or the position spec
            |                           # OR
            \s*                         # A line only containing white space
         )
        [\n]                            # line break at the end
    )+
)                                       # A positions block should be one or more lines
"""


class BlockIterator(Iterator):
    """
    A iterator for wrapping the iterator returned by `match.finditer`
    to extract the required fields directly from the match object
    """

    def __init__(self, it, natoms):
        self._it = it
        self._natoms = natoms
        self._catom = 0

    def __next__(self):
        try:
            match = next(self._it)
        except StopIteration:
            # if we reached the number of atoms declared, everything is well
            # and we re-raise the StopIteration exception
            if self._catom == self._natoms:
                raise
            else:
                # otherwise we got too less entries
                raise ValueError(f"Number of atom entries {self._catom} is smaller than the number of atoms {self._natoms}")

        self._catom += 1

        if self._catom > self._natoms:
            raise TypeError(f"Number of atom entries {self._catom} is larger than the number of atoms {self._natoms}")

        return (match["sym"], (float(match["x"]), float(match["y"]), float(match["z"])))


def parse_iter(fh_or_string):
    """Generates nested tuples for frames in XYZ files.

    Args:
        string: a string containing XYZ-structured text

    Yields:
        tuple: `(natoms, comment, atomiter)` for each frame
        in the XYZ data where `atomiter` is an iterator yielding a
        nested tuple `(symbol, (x, y, z))` for each entry.

    Raises:
        TypeError: If the number of atoms specified for the frame does not match
            the number of atom entries in the file.

    Examples:
        >>> print(len(list(parse_iter('''
        ... 5
        ... no comment
        ...  C         5.0000000000        5.0000000000        5.0000000000
        ...  H         5.6401052216        5.6401052216        5.6401052216
        ...  H         4.3598947806        4.3598947806        5.6401052208
        ...  H         4.3598947806        5.6401052208        4.3598947806
        ...  H         5.6401052208        4.3598947806        4.3598947806
        ... 5
        ... no comment
        ...  C         5.0000000000        5.0000000000        5.0000000000
        ...  H         5.6401902064        5.6401902064        5.6401902064
        ...  H         4.3598097942        4.3598097942        5.6401902063
        ...  H         4.3598097942        5.6401902063        4.3598097942
        ...  H         5.6401902063        4.3598097942        4.3598097942
        ... 5
        ... no comment
        ...  C         5.0000000000        5.0000000000        5.0000000000
        ...  H         5.6401902064        5.6401902064        5.6401902064
        ...  H         4.3598097942        4.3598097942        5.6401902063
        ...  H         4.3598097942        5.6401902063        4.3598097942
        ...  H         5.6401902063        4.3598097942        4.3598097942
        ... '''))))
        3
    """

    with as_byteorstringlike(fh_or_string) as (content, is_string):
        if is_string:
            frame_match = re.compile(FRAME_MATCH_REGEX, re.MULTILINE | re.VERBOSE)
            pos_match = re.compile(POS_MATCH_REGEX, re.MULTILINE | re.VERBOSE)
        else:
            frame_match = re.compile(FRAME_MATCH_REGEX.encode("utf8"), re.MULTILINE | re.VERBOSE)
            pos_match = re.compile(POS_MATCH_REGEX.encode("utf8"), re.MULTILINE | re.VERBOSE)

        for block in frame_match.finditer(content):
            natoms = int(block["natoms"])
            yield (
                natoms,
                block["comment"] if is_string else block["comment"].decode("utf8"),
                BlockIterator(pos_match.finditer(block["positions"]), natoms),
            )


def parse(fh_or_string):
    """
    The same as parse_iter(...) but instead of iterators, a list of nested dicts containing again
    a list for the 'atoms' key instead of another iterator are returned.
    """
    return [
        {"natoms": natoms, "comment": comment, "atoms": list(atomiter)} for (natoms, comment, atomiter) in parse_iter(fh_or_string)
    ]



import json
import os
import pathlib
import sys

import click

from .blocks.common import merged_spans, span_char_count
from .parser import parse_iter_blocks


@click.command()
@click.argument("fhandle", metavar="[FILE|-]", type=click.File(), default="-")
@click.option(
    "-f",
    "--format",
    "oformat",
    type=click.Choice(("json", "yaml", "highlight")),
    default="json",
    help="Output format (json or yaml are structure formats, highlight shows which lines of the output have been matched)",
)
@click.option(
    "--color",
    "color",
    type=click.Choice(("auto", "always")),
    default="auto",
    help="When to colorize output",
)
@click.option("-s", "--safe-keys", is_flag=True, help="generate 'safe' key names (e.g. without spaces, dashes, ..)")
@click.option("-S", "--statistics", is_flag=True, help="print some statistics to stderr")
@click.option("-k", "--key", "paths", metavar="<PATH>", type=str, multiple=True, help="Path, ex.: 'energies/total force_eval'")
def cp2kparse(fhandle, oformat, color, safe_keys, statistics, paths):
    """Parse the CP2K output FILE and return a structured output"""

    tree = {}
    spans = []

    content = fhandle.read()

    for match in parse_iter_blocks(content, key_mangling=safe_keys):
        tree.update(match.data)
        spans += match.spans

    spans = merged_spans(spans)

    def _(val):
        if isinstance(val, list):
            return ", ".join(str(v) for v in val)

        return val

    if paths:
        assert oformat != "parsed", "When selecting a specific path to output, json or yaml output format has to be used"
        for path in paths:
            sections = path.split("/")
            ref = tree
            for section in sections:
                if isinstance(ref, list):
                    section = int(section)  # if we encounter a list, convert the respective path element
                ref = ref[section]  # exploit Python using references into dicts/lists

            click.echo(f"{path}: {_(ref)}")

        return

    if oformat == "yaml":
        from ruamel.yaml import YAML

        yaml = YAML()
        yaml.dump(tree, sys.stdout)
    elif oformat == "highlight":
        ptr = 0
        for start, end in spans:
            click.secho(content[ptr:start], nl=False, dim=True, color=None if color == "auto" else True)
            click.secho(content[start:end], nl=False, bold=True, color=None if color == "auto" else True)
            ptr = end
        click.secho(content[ptr:], nl=False, dim=True, color=None if color == "auto" else True)

    else:
        click.echo(json.dumps(tree, indent=2, sort_keys=True))

    if statistics:
        click.echo("Statistics:\n===========\n", err=True)
        click.echo(f"Number of lines:      {len(content.splitlines()):>8}", err=True)
        click.echo(f"Number of characters: {len(content):>8}", err=True)
        click.echo(f"Percentage parsed:    {100*span_char_count(spans)/len(content):>8.2f}", err=True)


@click.command()
@click.argument("fhandle", metavar="[BANDSTRUCTURE_FILE|-]", type=click.File(), default="-")
@click.option(
    "-p",
    "--output-pattern",
    type=str,
    help="The output pattern for the different set files",
    default="{bsfile_basename}.set-{setnr}.csv",
    show_default=True,
)
@click.option(
    "-C",
    "--output-dir",
    type=click.Path(exists=True, file_okay=False, writable=True, path_type=pathlib.Path),
    help="Directory in which to create the CSV files",
    default=".",
    show_default=True,
)
def cp2k_bs2csv(fhandle, output_pattern, output_dir):
    """
    Convert the input from the given BANDSTRUCTURE_FILE (or stdin) and write
    CSV output files based on the given pattern.
    """
    from cp2k_output_tools.bandstructure_parser import set_gen

    content = fhandle.read()

    for setnr, npoints, _, specialpoints, points in set_gen(content):
        filename = output_pattern.format(bsfile_basename=os.path.basename(fhandle.name), setnr=setnr)

        print(f"writing point set {filename} (total number of k-points: {npoints})")
        with output_dir.joinpath(filename).open("w") as csvout:
            print("with the following special points:")

            for point in specialpoints:
                print(f"  {point.name:>8}: {point.a:10.8f} / {point.b:10.8f} / {point.c:10.8f}")

            for point in points:
                csvout.write(f"{point.a:10.8f} {point.b:10.8f} {point.c:10.8f}")
                for value in point.bands:
                    csvout.write(f" {value:10.8f}")
                csvout.write("\n")



from typing import Optional

import regex as re

from .common import FLOAT, BlockMatch

FORCES_RE = re.compile(
    rf"""
# anchor to indicate beginning of the FORCES output
^\ (?P<type>\w+)\ FORCES\ in\ \[(?P<unit>[^\]]+)\] \s* \n\n
 \ \#  .+ \n  # match the header
(
 \s+ (?P<atom>\d+) \s+ (?P<kind>\d+) \s+ (?P<element>\w+) \s+
     (?P<x>{FLOAT}) \s+ (?P<y>{FLOAT}) \s+ (?P<z>{FLOAT}) \s* \n
)+
\ SUM\ OF\ (?P<type>\w+)\ FORCES \s+
(
      (?P<sum_x>{FLOAT}) \s+
      (?P<sum_y>{FLOAT}) \s+
      (?P<sum_z>{FLOAT}) \s+
      (?P<norm>{FLOAT})
) \s* \n
""",
    re.VERSION1 | re.MULTILINE | re.VERBOSE,
)


def match_forces(content: str) -> Optional[BlockMatch]:
    forces = {}
    spans = []

    for match in FORCES_RE.finditer(content):
        ftype = match["type"].lower()

        per_atom = [
            {
                "atom": int(atom),
                "kind": int(kind),
                "element": element,
                "x": float(x),
                "y": float(y),
                "z": float(z),
            }
            for atom, kind, element, x, y, z in zip(*match.captures("atom", "kind", "element", "x", "y", "z"))
        ]
        fsum = {
            "x": float(match["sum_x"]),
            "y": float(match["sum_y"]),
            "z": float(match["sum_z"]),
            "norm": float(match["norm"]),
        }

        forces[ftype] = {"unit": match["unit"], "per_atom": per_atom, "sum": fsum}
        spans += match.spans(0)

    if not forces:
        return None

    return BlockMatch({"forces": forces}, spans)



from typing import Optional

import regex as re

from .common import FLOAT, BlockMatch

MULLIKEN_POPULATION_ANALYSIS_RE = re.compile(
    rf"""
# anchor to indicate beginning of the Mulliken Population Analysis
^[ \t]* Mulliken\ Population\ Analysis [ \t]* \n
 [ \t]* \n
 [ \t]* \#  [\w \t\,\(\)]+\n  # match the header
(
  ^
  [ \t]*
  (?P<atom>\d+) [ \t]+
  (?P<element>\w+) [ \t]+
  (?P<kind>\d+) [ \t]+
  (
    ( # spin unrestricted case:
      (?P<population_alpha>({FLOAT})) [ \t]+ (?P<population_beta>({FLOAT})) [ \t]+
      (?P<charge>({FLOAT})) [ \t]+ (?P<spin>({FLOAT}))
    )
    |
    (
      (?P<population>({FLOAT})) [ \t]+
      (?P<charge>({FLOAT}))
    )
  ) [ \t]*
  \n
)+
^ [ \t]* \#\ Total\ charge (\ and\ spin)? [ \t]+
(
  ( # spin unrestricted case:
    (?P<total_population_alpha>({FLOAT})) [ \t]+ (?P<total_population_beta>({FLOAT})) [ \t]+
    (?P<total_charge>({FLOAT})) [ \t]+ (?P<total_spin>({FLOAT}))
  )
  |
  (
    (?P<total_population>({FLOAT})) [ \t]+
    (?P<total_charge>({FLOAT}))
  )
) [ \t]*
\n
""",
    re.VERSION1 | re.MULTILINE | re.VERBOSE,
)


def match_mulliken_population_analysis(content: str) -> Optional[BlockMatch]:
    match = MULLIKEN_POPULATION_ANALYSIS_RE.search(content)

    if match is None:
        return None

    captures = match.capturesdict()
    per_atom = []

    if captures.get("population_alpha"):
        for idx in range(len(captures["atom"])):
            per_atom.append(
                {
                    "element": captures["element"][idx],
                    "kind": int(captures["kind"][idx]),
                    "population alpha": float(captures["population_alpha"][idx]),
                    "population beta": float(captures["population_beta"][idx]),
                    "charge": float(captures["charge"][idx]),
                    "spin": float(captures["spin"][idx]),
                }
            )

        return BlockMatch(
            {
                "mulliken population analysis": {
                    "per atom": per_atom,
                    "total": {
                        "population alpha": float(captures["total_population_alpha"][0]),
                        "population beta": float(captures["total_population_beta"][0]),
                        "charge": float(captures["total_charge"][0]),
                        "spin": float(captures["total_spin"][0]),
                    },
                }
            },
            match.spans(0),
        )

    # spin-restricted case:
    for idx in range(len(captures["atom"])):
        per_atom.append(
            {
                "element": captures["element"][idx],
                "kind": int(captures["kind"][idx]),
                "population": float(captures["population"][idx]),
                "charge": float(captures["charge"][idx]),
            }
        )

    return BlockMatch(
        {
            "mulliken population analysis": {
                "per atom": per_atom,
                "total": {"population": float(captures["total_population"][0]), "charge": float(captures["total_charge"][0])},
            }
        },
        match.spans(0),
    )



from typing import Optional

import regex as re

from .common import BlockMatch

DOI_RE = re.compile(r"^\s*\Khttps?://doi\.org/[^\s]+$", re.MULTILINE | re.VERSION1)


def match_bibliography(content: str) -> Optional[BlockMatch]:
    """Currently we only match the DOIs, but any decent bibliography manager can resolve this metadata"""

    bibliography = []
    spans = []

    for match in DOI_RE.finditer(content):
        bibliography.append({"doi": match[0]})
        spans += match.spans(0)

    if not bibliography:
        return None

    return BlockMatch({"bibliography": bibliography}, spans)



import argparse
import contextlib
import mmap
import re
import sys

from .xyz import FRAME_MATCH_REGEX


@contextlib.contextmanager
def mmapped(fhandle):
    fmapped = mmap.mmap(fhandle.fileno(), 0, access=mmap.ACCESS_READ)

    try:
        yield fmapped
    finally:
        fmapped.close()


CP2K_COMMENT_MATCH = re.compile(
    r"""
^[ \t]* i    [ ] = [ \t]+ (?P<iteration> \d+),
 [ \t]* time [ ] = [ \t]+ (?P<time> [\+\-]?  ( \d*[\.]\d+  | \d+[\.]?\d* )  ([Ee][\+\-]?\d+)? ),
 [ \t]* E    [ ] = [ \t]+ (?P<energy> [\+\-]?  ( \d*[\.]\d+  | \d+[\.]?\d* )  ([Ee][\+\-]?\d+)? )
""",
    re.VERBOSE,
)

# this one acts directly on the mmapped byte-like contnet:
FRAME_MATCH = re.compile(FRAME_MATCH_REGEX.encode("utf8"), re.VERBOSE | re.MULTILINE)


def xyz_restart_cleaner():
    parser = argparse.ArgumentParser(description="Clean a trajectory written by CP2K from restarts, discarding discontinued frames")
    parser.add_argument("source", metavar="<source>", type=argparse.FileType("rb"), help="source trajectory file")
    parser.add_argument(
        "output",
        metavar="<output>",
        nargs="?",
        type=argparse.FileType("wb"),
        default=sys.stdout.buffer,
        help="destination trajectory file",
    )

    args = parser.parse_args()

    iteration = -1
    frames_cache = []

    with mmapped(args.source) as content:

        for frame_match in FRAME_MATCH.finditer(content):
            comment_match = CP2K_COMMENT_MATCH.match(frame_match["comment"].decode("utf8"))

            last_iteration = iteration
            iteration = int(comment_match["iteration"])

            if iteration <= last_iteration:

                try:
                    ndrop = next(n for n, (i, f) in enumerate(reversed(frames_cache), start=1) if i < iteration) - 1
                except StopIteration:
                    print(
                        "WARNING: restart point may lie before already flushed frames, please run this again on the generated output",
                        file=sys.stderr,
                    )
                    ndrop = len(frames_cache)

                nflush = len(frames_cache) - ndrop
                print(f"found restart point @{last_iteration}, dropping {ndrop} frames, flushing {nflush}", file=sys.stderr)

                for frame in frames_cache[:-ndrop]:
                    args.output.write(frame[1])

                frames_cache.clear()

            frames_cache.append((iteration, frame_match[0]))

        print(f"flushing remaining {len(frames_cache)} frames", file=sys.stderr)

        for frame in frames_cache:
            args.output.write(frame[1])



from typing import Any, Dict, Optional

import regex as re

from .common import BlockMatch

KV_SECTION_RE = re.compile(
    r"""
^\ (
    (
        (?P<section>CP2K)\|\ (?P<key>cp2kflags)\: \s*(?P<value>.+?) \s* \n
        (?:\ CP2K\|\s{12} (?P<value>.+?) \s* \n)*
    )
    |
    (?P<section>\L<sections>)\|\ (?P<key>\s*\w.+?):?\ {2,}(?P<value>.+) \n
    )
""",
    re.VERSION1 | re.VERBOSE | re.MULTILINE,
    sections=["CP2K", "GLOBAL", "DBCSR", "DFT", "QS"],
)


def match_kv_sections(content: str) -> Optional[BlockMatch]:
    result: Dict[str, Any] = {}
    spans = []

    subsections = {"dft": ("cutoffs",), "qs": ("multi grid cutoff [a.u.]", "interaction thresholds")}

    for match in KV_SECTION_RE.finditer(content):
        section = match["section"].lower()

        if section not in result:
            result[section] = {}

        key = match["key"].lower()

        if section in subsections and any(key.startswith(k) for k in subsections[section]):
            subsection = next(k for k in subsections[section] if key.startswith(k))
            # if we encounter the cutoffs key in the DFT section we have subkes
            key = key.lstrip(f"{subsection}: ")  # strip the subsection from the key
            result[section][subsection] = {}  # and create a new dict for it
            rpointer = result[section][subsection]  # make sure this (and following values) go into this subdict
        elif key.startswith("   "):
            # if we have an indented keyword it belongs to the subsection started above
            key = key.lstrip()
        else:
            # otherwise we have regular keywords again
            rpointer = result[section]

        # for line continued values we get a list of captured values, rejoin them here
        value = "".join(match.captures("value"))
        spans += match.spans(0)

        if value in ["T", "F"]:
            rpointer[key] = True if value == "T" else False
            continue

        try:
            rpointer[key] = int(value)
            continue
        except ValueError:
            pass

        try:
            rpointer[key] = float(value)
            continue
        except ValueError:
            pass

        rpointer[key] = value

    if not result:
        return None

    return BlockMatch(result, spans)



from typing import Any, Dict, Optional

import regex as re

from .common import BlockMatch

WARNING_MESSAGE_RE = re.compile(
    r"""
^\ \*{3}\ WARNING\ in\ (?<filename>[^:]+):(?P<line>\d+)\ ::\ (?P<message>.+?) \*{3} \n
(\ \*{3}\ (?P<message>.+?) \*{3} \n)*
""",
    re.VERSION1 | re.VERBOSE | re.MULTILINE,
)

WARNING_MESSAGE_SIRIUS_RE = re.compile(
    r"""
^===\ (?P<type>Warning|Fatal\ error)\ at\ line\ (?P<line>\d+)\ of\ file\ (?P<filename>.+?)\ ===\n
(?P<message>.+) \n
(?:(?P<details>.+) \n)*
""",
    re.VERSION1 | re.VERBOSE | re.MULTILINE,
)

TOTAL_WARNING_COUNT_RE = re.compile(
    r"""
^\ The\ number\ of\ warnings\ for\ this\ run\ is\ :\s* (?P<value>\d+)
""",
    re.VERBOSE | re.MULTILINE,
)


def match_warnings(content: str) -> Optional[BlockMatch]:
    result: Dict[str, Any] = {"warnings": []}
    spans = []

    for wmatch in WARNING_MESSAGE_RE.finditer(content):
        result["warnings"] += [
            {"filename": wmatch["filename"], "line": int(wmatch["line"]), "message": "".join(wmatch.captures("message")).rstrip()}
        ]
        spans += wmatch.spans(0)

    match = TOTAL_WARNING_COUNT_RE.search(content)
    if match:
        result["nwarnings"] = int(match["value"])
        spans += match.spans(0)

    for wmatch in WARNING_MESSAGE_SIRIUS_RE.finditer(content):
        msg = {"filename": wmatch["filename"], "line": int(wmatch["line"]), "message": wmatch["message"].strip()}
        if wmatch.captures("details"):
            msg["details"] = [d.strip() for d in wmatch.captures("details")]

        if wmatch["type"] == "Warning":
            result["warnings"].append(msg)
            result.setdefault("nwarnings", 0)
            result["nwarnings"] += 1
        elif wmatch["type"] == "Fatal error":
            assert "error" not in result, "multiple fatal errors found when parsing SIRIUS warnings"
            result["error"] = msg
        else:
            raise AssertionError("invalid type found when parsing SIRIUS warnings")

        spans += wmatch.spans(0)

    return BlockMatch(result, spans)



from typing import Optional

import regex as re

from .common import FLOAT, BlockMatch

OVERLAP_MATRIX_CONDITION_NUMBER_RE = re.compile(
    rf"""
# anchor to indicate beginning the overlap matrix condition number section
^[ \t]* OVERLAP\ MATRIX\ CONDITION\ NUMBER\ AT\ GAMMA\ POINT [ \t]* \n
 [ \t]* 1-Norm\ Condition\ Number\ \(Estimate\) [ \t]* \n
 [ \t]* CN\ :\ \|A\|\*\|A\^-1\|:
   [ \t]* (?P<norm1_estimate_A>({FLOAT}))
   [ \t]* \*
   [ \t]* (?P<norm1_estimate_Ainv>({FLOAT}))
   [ \t]* =
   [ \t]* (?P<norm1_estimate>({FLOAT}))
   [ \t]* Log\(1-CN\):
   [ \t]* (?P<norm1_estimate_log>({FLOAT}))
   [ \t]* \n

 [ \t]* 1-Norm\ and\ 2-Norm\ Condition\ Numbers\ using\ Diagonalization [ \t]* \n

 [ \t]* CN\ :\ \|A\|\*\|A\^-1\|:
   [ \t]* (?P<norm1_diag_A>({FLOAT})) [ \t]* \* [ \t]* (?P<norm1_diag_Ainv>({FLOAT}))
   [ \t]* =
   [ \t]* (?P<norm1_diag>({FLOAT})) [ \t]* Log\(1-CN\): [ \t]* (?P<norm1_diag_log>({FLOAT}))
   [ \t]* \n

 [ \t]* CN\ :\ max/min\ ev:
   [ \t]* (?P<norm2_diag_max_ev>({FLOAT})) [ \t]* / [ \t]* (?P<norm2_diag_min_ev>({FLOAT}))
   [ \t]* =
   [ \t]* (?P<norm2_diag>({FLOAT})) [ \t]* Log\(2-CN\): [ \t]* (?P<norm2_diag_log>({FLOAT}))
   [ \t]* \n
""",
    re.MULTILINE | re.VERBOSE,
)


def match_overlap_matrix_condition_number(content: str) -> Optional[BlockMatch]:
    match = OVERLAP_MATRIX_CONDITION_NUMBER_RE.search(content)

    if match is None:
        return None

    captures = match.groupdict()

    return BlockMatch(
        {
            "overlap_matrix_condition_number": {
                "1-norm (estimate)": {
                    "|A|": float(captures["norm1_estimate_A"]),
                    "|A^-1|": float(captures["norm1_estimate_Ainv"]),
                    "CN": float(captures["norm1_estimate"]),
                    "Log(CN)": float(captures["norm1_estimate_log"]),
                },
                "1-norm (using diagonalization)": {
                    "|A|": float(captures["norm1_diag_A"]),
                    "|A^-1|": float(captures["norm1_diag_Ainv"]),
                    "CN": float(captures["norm1_diag"]),
                    "Log(CN)": float(captures["norm1_diag_log"]),
                },
                "2-norm (using diagonalization)": {
                    "max EV": float(captures["norm2_diag_max_ev"]),
                    "min EV": float(captures["norm2_diag_min_ev"]),
                    "CN": float(captures["norm2_diag"]),
                    "Log(CN)": float(captures["norm2_diag_log"]),
                },
            }
        },
        match.spans(0),
    )



from typing import Optional

import regex as re

from .common import FLOAT, BlockMatch

FORCE_EVAL_ENERGY_RE = re.compile(
    rf"""
^\s*ENERGY\|\ Total\ FORCE_EVAL [^:]+:\s*(?P<value>{FLOAT})\n
""",
    re.VERBOSE | re.MULTILINE,
)


def match_energies(content: str) -> Optional[BlockMatch]:
    match = FORCE_EVAL_ENERGY_RE.search(content)

    if not match:
        return None

    return BlockMatch({"energies": {"total force_eval": float(match["value"])}}, match.spans(0))



"""
Parse the CP2K bandstructure output file to a set of dataclasses
"""

__all__ = ["SpecialPoint", "Point", "BandstructureSet", "parse_bandstructure", "set_gen"]


import itertools
import re
from dataclasses import dataclass
from decimal import Decimal
from typing import Iterator, List, Optional, Tuple


@dataclass
class SpecialPoint:
    number: int
    name: str
    a: Decimal
    b: Decimal
    c: Decimal


@dataclass
class Point:
    a: Decimal
    b: Decimal
    c: Decimal
    bands: List[Decimal]
    spin: int
    weight: Optional[Decimal] = None


@dataclass
class BandstructureSet:
    setnr: int
    npoints: int
    nbands: Optional[int]
    points: List[Point]
    specialpoints: List[SpecialPoint]


SET_MATCH = re.compile(
    r"""
[ ]*
  SET: [ ]* (?P<setnr>\d+) [ ]*
  TOTAL [ ] POINTS: [ ]* (?P<npoints>\d+) [ ]*
  \n
(?P<content>
  [\s\S]*?(?=\n.*?[ ] SET|$)  # match everything until next 'SET' or EOL
)
""",
    re.VERBOSE,
)

SPOINTS_MATCH = re.compile(
    r"""
[ ]*
  POINT [ ]+ (?P<number>\d+) [ ]+ (?P<name>\S+) [ ]+ (?P<a>\S+) [ ]+ (?P<b>\S+) [ ]+ (?P<c>\S+)
""",
    re.VERBOSE,
)

POINTS_MATCH = re.compile(
    r"""
[ ]*
  Nr\. [ ]+ (?P<nr>\d+) [ ]+
  Spin [ ]+ (?P<spin>\d+) [ ]+
  K-Point [ ]+ (?P<a>\S+) [ ]+ (?P<b>\S+) [ ]+ (?P<c>\S+) [ ]*
  \n
[ ]* (?P<npoints>\d+) [ ]* \n
(?P<bands>
  [\s\S]*?(?=\n.*?[ ] Nr|$)  # match everything until next 'Nr.' or EOL
)
""",
    re.VERBOSE,
)


def _specialpoints_gen(content) -> Iterator[SpecialPoint]:
    for match in SPOINTS_MATCH.finditer(content):
        yield SpecialPoint(int(match["number"]), match["name"], Decimal(match["a"]), Decimal(match["b"]), Decimal(match["c"]))


def _points_gen(content) -> Iterator[Point]:
    for match in POINTS_MATCH.finditer(content):
        yield Point(
            a=Decimal(match["a"]),
            b=Decimal(match["b"]),
            c=Decimal(match["c"]),
            bands=[Decimal(v) for v in match["bands"].split()],
            spin=int(match["spin"]),
        )


SET_MATCH8 = re.compile(
    r"""
\#\ Set\ (?P<setnr>\d+):\ \d+\ special\ points,\ (?P<npoints>\d+)\ k-points,\ (?P<nbands>\d+)\ bands \s*
(?P<content>
  [\s\S]*?(?=\n.*?\#\ Set|$)  # match everything until next 'SET' or EOL
)
""",
    re.VERBOSE,
)


SPOINTS_MATCH8 = re.compile(
    r"""
\#\s+ Special\ point\ (?P<number>\d+) \s+ (?P<a>\S+) \s+ (?P<b>\S+) \s+ (?P<c>\S+) \s+ (?P<name>\S+)
""",
    re.VERBOSE,
)


POINTS_MATCH8 = re.compile(
    r"""
\#\ \ Point\ (?P<nr>\d+)\s+ Spin\ (?P<spin>\d+): \s+ (?P<a>\S+) \s+ (?P<b>\S+) \s+ (?P<c>\S+) [ ]* ((?P<weight>\S+) [ ]*)? \n
\#\ \ \ Band \s+ Energy\ \[eV\] \s+ Occupation \s*
(?P<bands>
  [\s\S]*?(?=\n.*?\#\ \ Point|$)  # match everything until next '# Point' or EOL
)
""",
    re.VERBOSE,
)


def _points_gen8(content: str) -> Iterator[Point]:
    for match in POINTS_MATCH8.finditer(content):
        weight: Optional[Decimal] = None
        try:
            weight = Decimal(match["weight"])
        except TypeError:
            pass  # ignore if None, conversion error is a ValueError

        values = match["bands"].split()

        yield Point(
            a=Decimal(match["a"]),
            b=Decimal(match["b"]),
            c=Decimal(match["c"]),
            bands=[Decimal(v) for v in values[1::3]],
            weight=weight,
            spin=int(match["spin"]),
        )


def _specialpoints_gen8(content: str) -> Iterator[SpecialPoint]:
    for match in SPOINTS_MATCH8.finditer(content):
        yield SpecialPoint(int(match["number"]), match["name"], Decimal(match["a"]), Decimal(match["b"]), Decimal(match["c"]))


def set_gen(content: str) -> Iterator[Tuple[int, int, Iterator[SpecialPoint], Iterator[Point]]]:
    """
    Parse the bandstructure from a CP2K bandstructure file content and return
    a tuple of metadata and SpecialPoint/Point Iterator.
    Supports both CP2K v8+ and previous version format transparently.
    """

    # try with the CP2K+8+ regex first
    matchiter = SET_MATCH8.finditer(content)
    specialpoints_gen = _specialpoints_gen8
    points_gen = _points_gen8

    try:
        peek = next(matchiter)
        matchiter = itertools.chain([peek], matchiter)
    except StopIteration:
        # if nothing could be found, fallback to the older format
        matchiter = SET_MATCH.finditer(content)
        specialpoints_gen = _specialpoints_gen
        points_gen = _points_gen

    for match in matchiter:
        nbands: Optional[int] = None

        try:
            nbands = int(match["nbands"])
        except IndexError:  # v7 doesn't have it (only implicitly)
            pass

        yield (
            int(match["setnr"]),
            int(match["npoints"]),
            nbands,
            specialpoints_gen(match["content"]),
            points_gen(match["content"]),
        )


def parse_bandstructure(content: str) -> Iterator[BandstructureSet]:
    """
    Parse the bandstructure from a CP2K bandstructure file content and return instances of BandstructureSet.
    """

    for setnr, npoints, nbands, spgen, pgen in set_gen(content):
        yield BandstructureSet(setnr, npoints, nbands, list(pgen), list(spgen))



from typing import Any, Callable, Dict, Iterator, List

from .blocks import builtin_matchers
from .blocks.common import BlockMatch, safe_string


def _mangled_keys(idict, mangle_func):
    if isinstance(idict, dict):
        return {mangle_func(k): _mangled_keys(v, mangle_func) for k, v in idict.items()}

    if isinstance(idict, list):
        return [_mangled_keys(i, mangle_func) for i in idict]

    return idict


def parse_iter_blocks(
    content: str, matchers: List[Callable] = builtin_matchers, key_mangling: bool = False
) -> Iterator[BlockMatch]:
    """Yield BlockMatch objects containing both structured data and metadata for each found match"""
    for matcher in matchers:
        match = matcher(content)

        if match:
            if key_mangling:
                yield BlockMatch(_mangled_keys(match.data, safe_string), match.spans)
            else:
                yield match


def parse_iter(content: str, matchers: List[Callable] = builtin_matchers, key_mangling: bool = False) -> Iterator[Dict[str, Any]]:
    """Yields the structured data found in the block matches"""
    for block in parse_iter_blocks(content, matchers, key_mangling):
        yield block.data



from dataclasses import dataclass
from typing import Any, Dict, List, Tuple, Union

# floating point regex
FLOAT = r"[\+\-]?(\d*[\.]\d+|\d+[\.]?\d*)([Ee][\+\-]?\d+)?"


# convert possibly problematic characters to underscores
def safe_string(string):
    return (
        string.replace(" ", "_")
        .replace("-", "_")
        .replace("/", "_")
        .replace("[", "_")
        .replace("]", "_")
        .replace(".", "_")
        .replace("|", "_")
        .replace("^", "_")
        .replace("(", "_")
        .replace(")", "_")
        .replace("{", "_")
        .replace("}", "_")
    )


@dataclass
class BlockMatch:
    data: Dict[str, Any]  # the parsed/structured data
    spans: List[Union[int, Tuple[int, int]]]  # start and end character indices of the matches


def merged_spans(spans: List[Tuple[int, int]]):
    merged = [(-1, -1)]

    for start, end in sorted(spans):
        if start > merged[-1][1]:  # if the new start is after the latest end, add a new span
            merged.append((start, end))
        else:  # if not, keep the current start and replace its end
            merged[-1] = (merged[-1][0], max(merged[-1][1], end))

    del merged[0]
    return merged


def span_char_count(spans: List[Tuple[int, int]]):
    return sum(end - start for start, end in spans)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cp2k_output_tools',
 'cp2k_output_tools.blocks',
 'cp2k_output_tools.scripts',
 'cp2k_output_tools.trajectories']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.1,<9.0.0', 'numpy>=1.19,<2.0', 'regex>=2021.4,<2022.0']

extras_require = \
{':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.7,<1'],
 'yaml': ['ruamel.yaml>=0.16.5,<0.18.0']}

entry_points = \
{'console_scripts': ['cp2k_bs2csv = cp2k_output_tools.cli:cp2k_bs2csv',
                     'cp2k_pdos = cp2k_output_tools.scripts.pdos:cp2k_pdos',
                     'cp2kparse = cp2k_output_tools.cli:cp2kparse',
                     'xyz_restart_cleaner = '
                     'cp2k_output_tools.trajectories.xyz_cli:xyz_restart_cleaner']}

setup_kwargs = {
    'name': 'cp2k-output-tools',
    'version': '0.5.0',
    'description': 'Python tools to handle CP2K output files',
    'long_description': '# cp2k-output-tools\n\n[![Build Status](https://github.com/cp2k/cp2k-output-tools/actions/workflows/test.yml/badge.svg)](https://github.com/cp2k/cp2k-output-tools/actions) [![codecov](https://codecov.io/gh/cp2k/cp2k-output-tools/branch/develop/graph/badge.svg)](https://codecov.io/gh/cp2k/cp2k-output-tools) [![PyPI](https://img.shields.io/pypi/pyversions/cp2k-output-tools)](https://pypi.org/project/cp2k-output-tools/)\n\nModular CP2K output file parsers, mostly in the form of regular expressions plus other tools to mangle various CP2K output:\n\n  * `cp2kparse` ... parse CP2K output (for restart & input files look at the [cp2k-input-tools](https://github.com/cp2k/cp2k-input-tools) project) and allow easy selection of common values.\n  * `xyz_restart_parser` ... when restarts occur during an MD you may end up with duplicated frames in the trajectory, this tool filters them (and can easily handle huge files)\n  * `cp2k_bs2csv` ... convert a CP2K band structure file to multiple (one-per-set) CSV files for easier plotting. There is also an API available if you need to import bandstructure data into your application.\n  * `cp2k_pdos` ... apply a convolution with Gaussians on a regular grid on the CP2K PDOS output and generate a CSV file for further processing or plotting. The same grid is used for all input files with the min/max of the grid automatically determined, but no summation of the different projections is done.\n\n## Requirements\n\n* Python 3.6+\n* regex 2021+\n* click 8+\n* numpy 1.19+\n* optional: ruamel.yaml\n\nFor development: https://poetry.eustace.io/ https://pytest.org/\n\n\n## Usage: cp2kparse\n\nThere is a simple command-line interface `cp2kparse`:\n\n```console\n$ cp2kparse --help\nUsage: cp2kparse [OPTIONS] [FILE|-]\n\n  Parse the CP2K output FILE and return a structured output\n\nOptions:\n  -f, --format [json|yaml|highlight]\n                                  Output format (json or yaml are structure\n                                  formats, highlight shows which lines of the\n                                  output have been matched)\n  --color [auto|always]           When to colorize output\n  -s, --safe-keys                 generate \'safe\' key names (e.g. without\n                                  spaces, dashes, ..)\n  -S, --statistics                print some statistics to stderr\n  -k, --key <PATH>                Path, ex.: \'energies/total force_eval\'\n  --help                          Show this message and exit.\n\n$ cp2kparse calc.out\n{\n  "cp2k": {\n    "cp2kflags: libint fftw3 libxc xsmm spglib": " ",\n    "data directory path": "/data/tiziano/cp2k/data",\n    "input file name": "Si-supercell-001.inp",\n    "is freely available from": "https://www.cp2k.org/",\n    "program compiled at": "Mon Sep 16 14:42:43 CEST 2019",\n    "program compiled for": "local",\n    "program compiled on": "tcpc18",\n    "source code revision number": "git:43f50e2",\n    "version string": "CP2K version 7.0 (Development Version)"\n  },\n  "dbcsr": {\n    "cpu multiplication driver": "XSMM",\n    "maximum elements for images": "UNLIMITED",\n    "multiplication size stacks": 3,\n    "multiplication stack size": 1000,\n    "multiplicative factor virtual images": 1,\n    "multrec recursion limit": 512,\n    "use multiplication densification": true\n  },\n  "energies": {\n    "total force_eval": -251.6873903110507\n  },\n  "global": {\n    "all-to-all communication in single precision": false,\n    "basis set file name": "BASIS_MOLOPT",\n    "coordinate file name": "__STD_INPUT__",\n    "cpu model name": "Intel(R) Core(TM) i7-4790 CPU @ 3.60GHz",\n    "cpuid": 1002,\n    "ffts using library dependent lengths": false,\n    "force environment number": 1,\n    "global print level": "MEDIUM",\n    "method name": "CP2K",\n    "mm potential file name": "MM_POTENTIAL",\n    "mpi i/o enabled": true,\n    "number of threads for this process": 1,\n    "potential file name": "POTENTIAL",\n    "preferred diagonalization lib.": "SL",\n    "preferred fft library": "FFTW3",\n    "project name": "Si-supercell-001",\n    "run type": "ENERGY_FORCE",\n    "this output is from process": 0,\n    "total number of message passing processes": 1\n  },\n  "mulliken population analysis": {\n    "per atom": [\n      {\n        "charge": 7e-06,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.999993\n      },\n      {\n        "charge": 0.0,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.0\n      },\n      {\n        "charge": -0.0,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.0\n      },\n      {\n        "charge": -1e-06,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.000001\n      },\n      {\n        "charge": -0.0,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.0\n      },\n      {\n        "charge": -1e-06,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.000001\n      },\n      {\n        "charge": -1e-06,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.000001\n      },\n      {\n        "charge": -0.0,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.0\n      },\n      {\n        "charge": 0.000137,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.999863\n      },\n      {\n        "charge": -2e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.00002\n      },\n      {\n        "charge": -0.000133,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.000133\n      },\n      {\n        "charge": 2.1e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.999979\n      },\n      {\n        "charge": -0.000133,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.000133\n      },\n      {\n        "charge": 2.1e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.999979\n      },\n      {\n        "charge": 0.000137,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.999863\n      },\n      {\n        "charge": -2e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.00002\n      },\n      {\n        "charge": -3.9e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.000039\n      },\n      {\n        "charge": 3.6e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.999964\n      },\n      {\n        "charge": -1e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.00001\n      },\n      {\n        "charge": 1e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.99999\n      },\n      {\n        "charge": -3.9e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.000039\n      },\n      {\n        "charge": 3.6e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.999964\n      },\n      {\n        "charge": -1e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.00001\n      },\n      {\n        "charge": 1e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.99999\n      },\n      {\n        "charge": -3.9e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.000039\n      },\n      {\n        "charge": 3.6e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.999964\n      },\n      {\n        "charge": -3.9e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.000039\n      },\n      {\n        "charge": 3.6e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.999964\n      },\n      {\n        "charge": -1e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.00001\n      },\n      {\n        "charge": 1e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.99999\n      },\n      {\n        "charge": -1e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.00001\n      },\n      {\n        "charge": 1e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.99999\n      },\n      {\n        "charge": 6e-06,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.999994\n      },\n      {\n        "charge": 3.5e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.999965\n      },\n      {\n        "charge": 3.3e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.999967\n      },\n      {\n        "charge": 9e-06,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.999991\n      },\n      {\n        "charge": 3.3e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.999967\n      },\n      {\n        "charge": 9e-06,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.999991\n      },\n      {\n        "charge": 3e-06,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.999997\n      },\n      {\n        "charge": -0.000312,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.000312\n      },\n      {\n        "charge": 3e-06,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.999997\n      },\n      {\n        "charge": -0.000311,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.000311\n      },\n      {\n        "charge": 3.4e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.999966\n      },\n      {\n        "charge": 1e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.99999\n      },\n      {\n        "charge": 3.4e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.999966\n      },\n      {\n        "charge": 1e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.99999\n      },\n      {\n        "charge": 6e-06,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.999994\n      },\n      {\n        "charge": 3.5e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.999965\n      },\n      {\n        "charge": -9e-06,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.000009\n      },\n      {\n        "charge": -3.3e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.000033\n      },\n      {\n        "charge": 0.000308,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.999692\n      },\n      {\n        "charge": -4e-06,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.000004\n      },\n      {\n        "charge": -3.6e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.000036\n      },\n      {\n        "charge": -7e-06,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.000007\n      },\n      {\n        "charge": -9e-06,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.000009\n      },\n      {\n        "charge": -3.3e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.000033\n      },\n      {\n        "charge": -8e-06,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.000008\n      },\n      {\n        "charge": -3.3e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.000033\n      },\n      {\n        "charge": -3.5e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.000035\n      },\n      {\n        "charge": -6e-06,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.000006\n      },\n      {\n        "charge": 0.000308,\n        "element": "Si",\n        "kind": 1,\n        "population": 3.999692\n      },\n      {\n        "charge": -4e-06,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.000004\n      },\n      {\n        "charge": -8e-06,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.000008\n      },\n      {\n        "charge": -3.3e-05,\n        "element": "Si",\n        "kind": 1,\n        "population": 4.000033\n      }\n    ],\n    "total": {\n      "charge": 0.0,\n      "population": 256.0\n    }\n  },\n  "program info": {\n    "ended at": "2019-09-17 17:02:35.738",\n    "process id": "23773",\n    "ran by": "tiziano",\n    "ran on": "tcpc18",\n    "started at": "2019-09-17 17:01:19.273",\n    "started by": "tiziano",\n    "started in": "/users/tiziano/work/phonopy/example/Si-CP2K",\n    "started on": "tcpc18",\n    "stopped in": "/users/tiziano/work/phonopy/example/Si-CP2K"\n  }\n}\n```\n\nand an API:\n\n```python\nfrom cp2k_output_tools import parse_iter\n\nwith open("calc.out", "r") as fhandle:\n    for match in parse_iter(fhandle.read()):\n        print(match.values)\n```\n\n## Usage: xyz_restart_cleaner\n\n```console\n$ xyz_restart_cleaner orig_trajectory.xyz new_trajectory.xyz\nfound restart point @1, dropping 1 frames, flushing 1\nflushing remaining 2 frames\n```\n\n## Usage: cp2k_bs2csv\n\nGiven a `.bs` file as written by CP2K the script will generate CSV files with the same name and suffixes `.set-X.csv` in the same directory as the source file:\n\n```console\n$ cp2k_bs2csv WO3.bs\nwriting point set WO3.bs.set-1.csv (total number of k-points: 11)\nwith the following special points:\n     GAMMA: 0.00000000 / 0.00000000 / 0.00000000\n         X: 0.00000000 / 0.50000000 / 0.00000000\nwriting point set WO3.bs.set-2.csv (total number of k-points: 11)\nwith the following special points:\n         X: 0.00000000 / 0.50000000 / 0.00000000\n         M: 0.50000000 / 0.50000000 / 0.00000000\nwriting point set WO3.bsset-3.csv (total number of k-points: 11)\nwith the following special points:\n         M: 0.50000000 / 0.50000000 / 0.00000000\n     GAMMA: 0.00000000 / 0.00000000 / 0.00000000\nwriting point set WO3.bs.set-4.csv (total number of k-points: 11)\nwith the following special points:\n     GAMMA: 0.00000000 / 0.00000000 / 0.00000000\n         R: 0.50000000 / 0.50000000 / 0.50000000\nwriting point set WO3.bs.set-5.csv (total number of k-points: 11)\nwith the following special points:\n         R: 0.50000000 / 0.50000000 / 0.50000000\n         X: 0.00000000 / 0.50000000 / 0.00000000\nwriting point set WO3.bs.set-6.csv (total number of k-points: 11)\nwith the following special points:\n         R: 0.50000000 / 0.50000000 / 0.50000000\n         M: 0.50000000 / 0.50000000 / 0.00000000\n$ ls\nWO3.cp2k-8.bs            WO3.cp2k-8.bs.set-2.csv  WO3.cp2k-8.bs.set-4.csv  WO3.cp2k-8.bs.set-6.csv\nWO3.cp2k-8.bs.set-1.csv  WO3.cp2k-8.bs.set-3.csv  WO3.cp2k-8.bs.set-5.csv\n```\n\n## Development\n\n```console\n$ poetry install\n$ poetry run pytest -v\n```\n',
    'author': 'Tiziano Müller',
    'author_email': 'tiziano.mueller@chem.uzh.ch',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/cp2k/cp2k-output-tools',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""
Convert the discrete CP2K PDOS points to a smoothed curve using convoluted gaussians.

If you have separate alpha/beta-spin pdos files (spin-unrestricted calculation),
pass both files as arguments to get one common grid for both of them.
"""

# Copyright (c) 2020 Tiziano Müller <tiziano.mueller@chem.uzh.ch>,
# based on a Fortran tool written by Marcella Iannuzzi <marcella.iannuzzi@chem.uzh.ch>


import argparse
import contextlib
import re
import sys

import numpy as np

HEADER_MATCH = re.compile(
    r"\# Projected DOS for (?:atomic kind \w+|list \d+ of \d+ atoms,)"
    r" at iteration step i = \d+, E\(Fermi\) = [ \t]* (?P<Efermi>[^\t ]+) a\.u\."
)

# Column indexes, starting from 0
EIGENVALUE_COLUMN = 1
DENSITY_COLUMN = 3


# from https://stackoverflow.com/a/17603000/1400465
@contextlib.contextmanager
def smart_open(filename=None):
    if filename and filename != "-":
        fhandle = open(filename, "w")
    else:
        fhandle = sys.stdout

    try:
        yield fhandle
    finally:
        if fhandle is not sys.stdout:
            fhandle.close()


def cp2k_pdos():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "pdosfilenames",
        metavar="<PDOS-file>",
        type=str,
        nargs="+",
        help="the PDOS file generated by CP2K, specify two files (alpha/beta) in a spin-unrestricted case",
    )
    parser.add_argument("--sigma", "-s", type=float, default=0.02, help="sigma for the gaussian distribution (default: 0.02)")
    parser.add_argument("--de", "-d", type=float, default=0.001, help="integration step size (default: 0.001)")
    parser.add_argument("--scale", "-c", type=float, default=1, help="scale the density by this factor (default: 1)")
    parser.add_argument("--total-sum", action="store_true", help="calculate and print the total sum for each orbital (default: no)")
    parser.add_argument("--no-header", action="store_true", default=False, help="do not print a header (default: print header)")
    parser.add_argument(
        "--output", "-o", type=str, default=None, help="write output to specified file (default: write to standard output)"
    )
    args = parser.parse_args()

    alldata = []
    orb_headers = []

    for pdosfilename in args.pdosfilenames:
        with open(pdosfilename, "r") as fhandle:
            match = HEADER_MATCH.match(fhandle.readline().rstrip())
            if not match:
                print(
                    f"The file '{pdosfilename}' does not look like a CP2K PDOS output.\n"
                    "If it is indeed a correct output file, please report an issue at\n"
                    "    https://github.com/cp2k/cp2k-output-tools/issues"
                )
                sys.exit(1)

            efermi = float(match.group("Efermi"))
            # header is originally: ['#', 'MO', 'Eigenvalue', '[a.u.]', 'Occupation', 's', 'py', ...]
            header = fhandle.readline().rstrip().split()[1:]  # remove the comment directly
            header[1:3] = [" ".join(header[1:3])]  # rejoin "Eigenvalue" and its unit
            data = np.loadtxt(fhandle)  # load the rest directly with numpy

        alldata.append(data)

        orb_headers += header[DENSITY_COLUMN:]

    # take the boundaries over all energy eigenvalues (not guaranteed to be the same)
    # add a margin to not cut-off Gaussians at the borders
    margin = 10 * args.sigma
    emin = min(np.min(data[:, EIGENVALUE_COLUMN]) for data in alldata) - margin
    emax = max(np.max(data[:, EIGENVALUE_COLUMN]) for data in alldata) + margin
    ncols = sum(data.shape[1] - DENSITY_COLUMN for data in alldata)
    nmesh = int((emax - emin) / args.de) + 1  # calculate manually instead of using np.arange to ensure emax inside the mesh
    xmesh = np.linspace(emin, emax, nmesh)
    ymesh = np.zeros((nmesh, ncols))

    # printing to stderr makes it possible to simply redirect the stdout to a file
    print(
        f"Integration step:  {args.de:14.5f}\n"
        f"Sigma:             {args.sigma:14.5f}\n"
        f"Minimum energy:    {emin+margin:14.5f} - {margin:.5f}\n"
        f"Maximum energy:    {emax-margin:14.5f} + {margin:.5f}\n"
        f"Nr of mesh points: {nmesh:14d}",
        file=sys.stderr,
    )

    fact = args.de / (args.sigma * np.sqrt(2.0 * np.pi))

    coloffset = 0
    for fname, data in zip(args.pdosfilenames, alldata):
        print(f"Nr of lines:       {data.shape[0]:14d} in {fname}", file=sys.stderr)
        ncol = data.shape[1] - DENSITY_COLUMN

        for idx in range(nmesh):
            func = np.exp(-((xmesh[idx] - data[:, EIGENVALUE_COLUMN]) ** 2) / (2.0 * args.sigma ** 2)) * fact
            ymesh[idx, coloffset : (coloffset + ncol)] = func.dot(data[:, DENSITY_COLUMN:])

        coloffset += ncol

    if args.total_sum:
        finalsum = np.sum(ymesh, 0) * args.de
        print("Sum over all meshpoints, per orbital:", file=sys.stderr)
        print(("{:16.8f}" * ncols).format(*finalsum), file=sys.stderr)

    xmesh -= efermi  # put the Fermi energy at 0
    xmesh *= 27.211384  # convert to eV
    ymesh *= args.scale  # scale

    with smart_open(args.output) as fhandle:
        if not args.no_header:
            print(("{:>16}" + " {:>16}" * ncols).format("Energy_[eV]", *orb_headers), file=fhandle)
        for idx in range(nmesh):
            print(("{:16.8f}" + " {:16.8f}" * ncols).format(xmesh[idx], *ymesh[idx, :]), file=fhandle)


#  vim: set ts=4 sw=4 tw=0 :



from .bibliography import match_bibliography
from .condition_number import match_overlap_matrix_condition_number
from .energies import match_energies
from .forces import match_forces
from .kv_sections import match_kv_sections
from .mulliken import match_mulliken_population_analysis
from .program_info import match_program_info
from .warnings import match_warnings

builtin_matchers = [
    match_overlap_matrix_condition_number,
    match_mulliken_population_analysis,
    match_program_info,
    match_kv_sections,
    match_energies,
    match_forces,
    match_warnings,
    match_bibliography,
]
